/*******************************************************************************
 
  $Id: console_io.cpp,v 1.1 1999-08-22 11:03:32+04 eug Exp eug $

  (C) Copyright 1998-99 by Eugene Surovegin <ebs@innocent.com>

  Console I/O

*******************************************************************************/
#include <windows.h>
#include <malloc.h>
#include <stdarg.h>

#include "console_io.h"

TConsoleIO::TConsoleIO(){ Init(); }

TConsoleIO::TConsoleIO(const char* title){
  Init();
  SetTitle(title);
}

void TConsoleIO::Init(){

  std_in  = ::GetStdHandle(STD_INPUT_HANDLE);
  std_out = ::GetStdHandle(STD_OUTPUT_HANDLE);
  std_err = ::GetStdHandle(STD_ERROR_HANDLE);

  CONSOLE_SCREEN_BUFFER_INFO csbi;
  ::GetConsoleScreenBufferInfo(std_out, &csbi);
  save_attr = (BYTE)csbi.wAttributes;
}
    
TConsoleIO::~TConsoleIO(){
  RestoreTextAttr();
}

void TConsoleIO::SetTextAttribute(unsigned char attr){
  ::SetConsoleTextAttribute(std_out, attr); 
}

void TConsoleIO::SetTitle(const char* title){
  ::SetConsoleTitle(title);
}

void TConsoleIO::WriteFormattedString(const char* str){

  static char esc_delim[] ="\027";
  
  char* buf = (char*)alloca(::lstrlen(str)+1);
  ::lstrcpy(buf, str);

  RestoreTextAttr();

  char* token = strtok(buf, esc_delim);
  while (token != 0){
    if (token != buf){
      SetTextAttribute(*token);
      ++token;
    }

    WriteString(token);
    token = strtok(0, esc_delim);
  };
}

void TConsoleIO::PrintfString(const char* format, ...){
  char buf[2048];

  va_list l;
  va_start(l, format);
  ::wvsprintf(buf, format, l);
  va_end(l);

  WriteString(buf);
}

void TConsoleIO::PrintfFormattedString(const char* format, ...){
  char buf[2048];

  va_list l;
  va_start(l, format);
  ::wvsprintf(buf, format, l);
  va_end(l);
  WriteFormattedString(buf);
}

void TConsoleIO::WriteString(const char* str){
  DWORD __dummy;
  ::WriteConsole(std_out, str, ::lstrlen(str), &__dummy,  NULL);
}

void TConsoleIO::RestoreTextAttr(){
  SetTextAttribute(save_attr);
};