/*******************************************************************************
 
  $Id: main.cpp,v 1.1 1999-08-22 11:03:32+04 eug Exp eug $

  (C) Copyright 1999-2000 by Eugene Surovegin <ebs@innocent.com>
  Portions (C) Copyright by Microsoft Corp
  Portions (C) Copyright by Ralph Brown

  Thanks to "Giuz" <Giuseppe.DeVincenti@OnStream.com> for W2K fix

  Main module

*******************************************************************************/
#include <windows.h>
#include "console_io.h"
#include "build.h"

// Global console object  
static TConsoleIO* pc = 0;

// Ripped from NT DDK sources
#define FILE_DEVICE_DISK     0x00000007
#define IOCTL_DISK_BASE      FILE_DEVICE_DISK
#define FILE_READ_ACCESS     ( 0x0001 )    // file & pipe
#define FILE_WRITE_ACCESS    ( 0x0002 )    // file & pipe
#define IDENTIFY_BUFFER_SIZE 512

#define CTL_CODE( DeviceType, Function, Method, Access ) (                 \
    ((DeviceType) << 16) | ((Access) << 14) | ((Function) << 2) | (Method) \
)
#define METHOD_BUFFERED                 0
#define SMART_RCV_DRIVE_DATA CTL_CODE(IOCTL_DISK_BASE, 0x0022, METHOD_BUFFERED, FILE_READ_ACCESS | FILE_WRITE_ACCESS)

#define FILE_DEVICE_SCSI                0x0000001b
#define IOCTL_SCSI_MINIPORT_IDENTIFY    ((FILE_DEVICE_SCSI << 16) + 0x0501)
#define IOCTL_SCSI_MINIPORT 0x0004D008   // see NTDDSCSI.H for definition

#pragma pack(1)
typedef struct _IDEREGS {
        UCHAR    bFeaturesReg;           // Used for specifying SMART "commands".
        UCHAR    bSectorCountReg;        // IDE sector count register
        UCHAR    bSectorNumberReg;       // IDE sector number register
        UCHAR    bCylLowReg;             // IDE low order cylinder value
        UCHAR    bCylHighReg;            // IDE high order cylinder value
        UCHAR    bDriveHeadReg;          // IDE drive/head register
        UCHAR    bCommandReg;            // Actual IDE command.
        UCHAR    bReserved;              // reserved for future use.  Must be zero.
} IDEREGS, *PIDEREGS, *LPIDEREGS;

typedef struct _SENDCMDINPARAMS {
        ULONG   cBufferSize;            // Buffer size in bytes
        IDEREGS irDriveRegs;            // Structure with drive register values.
        UCHAR    bDriveNumber;          // Physical drive number to send
                                        // command to (0,1,2,3).
        UCHAR    bReserved[3];          // Reserved for future expansion.
        ULONG   dwReserved[4];          // For future use.
        UCHAR    bBuffer[1];            // Input buffer.
} SENDCMDINPARAMS, *PSENDCMDINPARAMS, *LPSENDCMDINPARAMS;

#define ID_CMD          0xEC            // Returns ID sector for ATA.

typedef struct _DRIVERSTATUS {
        UCHAR    bDriverError;          // Error code from driver,
                                        // or 0 if no error.
        UCHAR    bIDEError;             // Contents of IDE Error register.
                                        // Only valid when bDriverError
                                        // is SMART_IDE_ERROR.
        UCHAR    bReserved[2];          // Reserved for future expansion.
        ULONG   dwReserved[2];          // Reserved for future expansion.
} DRIVERSTATUS, *PDRIVERSTATUS, *LPDRIVERSTATUS;

typedef struct _SENDCMDOUTPARAMS {
        ULONG                   cBufferSize;            // Size of bBuffer in bytes
        DRIVERSTATUS            DriverStatus;           // Driver status structure.
        UCHAR                   bBuffer[1];             // Buffer of arbitrary length in which to store the data read from the
                                                        // drive.
} SENDCMDOUTPARAMS, *PSENDCMDOUTPARAMS, *LPSENDCMDOUTPARAMS;

typedef struct _IDENTIFY_DATA {
    USHORT GeneralConfiguration;            // 00 00
    USHORT NumberOfCylinders;               // 02  1
    USHORT Reserved1;                       // 04  2
    USHORT NumberOfHeads;                   // 06  3
    USHORT UnformattedBytesPerTrack;        // 08  4
    USHORT UnformattedBytesPerSector;       // 0A  5
    USHORT SectorsPerTrack;                 // 0C  6
    USHORT VendorUnique1[3];                // 0E  7-9
    USHORT SerialNumber[10];                // 14  10-19
    USHORT BufferType;                      // 28  20
    USHORT BufferSectorSize;                // 2A  21
    USHORT NumberOfEccBytes;                // 2C  22
    USHORT FirmwareRevision[4];             // 2E  23-26
    USHORT ModelNumber[20];                 // 36  27-46
    UCHAR  MaximumBlockTransfer;            // 5E  47
    UCHAR  VendorUnique2;                   // 5F
    USHORT DoubleWordIo;                    // 60  48
    USHORT Capabilities;                    // 62  49
    USHORT Reserved2;                       // 64  50
    UCHAR  VendorUnique3;                   // 66  51
    UCHAR  PioCycleTimingMode;              // 67
    UCHAR  VendorUnique4;                   // 68  52
    UCHAR  DmaCycleTimingMode;              // 69
    USHORT TranslationFieldsValid:1;        // 6A  53
    USHORT Reserved3:15;
    USHORT NumberOfCurrentCylinders;        // 6C  54
    USHORT NumberOfCurrentHeads;            // 6E  55
    USHORT CurrentSectorsPerTrack;          // 70  56
    ULONG  CurrentSectorCapacity;           // 72  57-58
    USHORT CurrentMultiSectorSetting;       //     59
    ULONG  UserAddressableSectors;          //     60-61
    USHORT SingleWordDMASupport : 8;        //     62
    USHORT SingleWordDMAActive : 8;
    USHORT MultiWordDMASupport : 8;         //     63
    USHORT MultiWordDMAActive : 8;
    USHORT AdvancedPIOModes : 8;            //     64
    USHORT Reserved4 : 8;
    USHORT MinimumMWXferCycleTime;          //     65
    USHORT RecommendedMWXferCycleTime;      //     66
    USHORT MinimumPIOCycleTime;             //     67
    USHORT MinimumPIOCycleTimeIORDY;        //     68
    USHORT Reserved5[2];                    //     69-70
    USHORT ReleaseTimeOverlapped;           //     71
    USHORT ReleaseTimeServiceCommand;       //     72
    USHORT MajorRevision;                   //     73
    USHORT MinorRevision;                   //     74
    USHORT Reserved6[50];                   //     75-126
    USHORT SpecialFunctionsEnabled;         //     127
    USHORT Reserved7[128];                  //     128-255
} IDENTIFY_DATA, *PIDENTIFY_DATA;

typedef struct _SRB_IO_CONTROL {
  ULONG HeaderLength;
  UCHAR Signature[8];
  ULONG Timeout;
  ULONG ControlCode;
  ULONG ReturnCode;
  ULONG Length;
} SRB_IO_CONTROL, *PSRB_IO_CONTROL;

#pragma pack()
// end of NT DDK rip

void print_ata_id(IDENTIFY_DATA* pId){

  bool ata = ((pId->GeneralConfiguration) & (1<<15)) == 0;
  
  pc->PrintfFormattedString("\n");
  if (ata){
    pc->PrintfFormattedString(LIGHTGREEN_B"ATA device (%s)\n", 
                              pId->GeneralConfiguration & (1<<6) ? "fixed" : "removable");
  }
  else {
    char type = (pId->GeneralConfiguration>>8) & 0x1f;
    static char* atapi_types[] = {
      "direct-access", 
      "sequential-access",
      "printer",
      "processor",
      "write-once",
      "CD-ROM",
      "scanner",
      "optical memory",
      "medium changer",
      "communications",
      "reserved for ACS IT8",
      "reserved for ACS IT8",
      "array controller",
      "unknown"
    };
    if (type > 0xd) type = 0xd;
    
    pc->PrintfFormattedString(LIGHTGREEN_B"ATAPI %s device (%s)\n", atapi_types[type], 
                              pId->GeneralConfiguration & (1<<7) ? "removable" : "fixed");
  }
  
  if (pId->ModelNumber[0]){
    pc->PrintfFormattedString("  Model................ ");
    char* p = (char*)pId->ModelNumber;
    for (int i=0; i < sizeof(pId->ModelNumber); i+=2)
      pc->PrintfFormattedString(YELLOW_B"%c%c", p[i+1], p[i]);
  }
  
  if (pId->FirmwareRevision[0]){
    pc->PrintfFormattedString("\n  Firmware revision.... ");
    char* p = (char*)pId->FirmwareRevision;
    for (int i=0; i < sizeof(pId->FirmwareRevision); i+=2)
      pc->PrintfFormattedString(YELLOW_B"%c%c", p[i+1], p[i]);
  }

  if (pId->SerialNumber[0]){
    pc->PrintfFormattedString("\n  Serial number........ ");
    char* p = (char*)pId->SerialNumber;
    for (int i=0; i < sizeof(pId->SerialNumber); i+=2)
      pc->PrintfFormattedString(YELLOW_B"%c%c", p[i+1], p[i]);
  }

  if (ata){
    if (pId->BufferSectorSize)
      pc->PrintfFormattedString("\n  Buffer size.......... "WHITE_B"%u (%u sectors)", 
                                    pId->BufferSectorSize*512, pId->BufferSectorSize);  

    pc->PrintfFormattedString("\n  Number of cylinders.. "WHITE_B"%u\n", pId->NumberOfCylinders);  
    pc->PrintfFormattedString("  Number of heads...... "WHITE_B"%u\n", pId->NumberOfHeads);
    pc->PrintfFormattedString("  Sectors per track.... "WHITE_B"%u", pId->SectorsPerTrack);  
  }
  pc->PrintfFormattedString("\n");
}

bool try_scsi_drive(int i){
  bool res = true;
  char dev_file[40];
  ::wsprintf(dev_file, "\\\\.\\Scsi%d:", i);

  #define LENGTH (sizeof(SENDCMDOUTPARAMS) + IDENTIFY_BUFFER_SIZE)

  char buffer[sizeof(SRB_IO_CONTROL) + LENGTH];
  SRB_IO_CONTROL* p = (SRB_IO_CONTROL*) buffer;
  SENDCMDINPARAMS* pin = (SENDCMDINPARAMS*)(buffer + sizeof(SRB_IO_CONTROL));

  DWORD dummy;
  HANDLE h = ::CreateFile(dev_file, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);
  if (h != INVALID_HANDLE_VALUE){
    for (int d = 0; d<2; ++d){
      ::ZeroMemory(buffer, sizeof(buffer));
      p->HeaderLength = sizeof(SRB_IO_CONTROL);
      p->Timeout = 10000;
      p->Length = LENGTH;
      p->ControlCode = IOCTL_SCSI_MINIPORT_IDENTIFY;
      ::CopyMemory(p->Signature, "SCSIDISK", 8);
  
      pin->irDriveRegs.bCommandReg = ID_CMD;
      pin->bDriveNumber = d;

      if (::DeviceIoControl(h, IOCTL_SCSI_MINIPORT, 
                            buffer, sizeof(SRB_IO_CONTROL) + sizeof(SENDCMDINPARAMS) - 1, 
                            buffer,  sizeof(SRB_IO_CONTROL) + LENGTH, &dummy, NULL)){

        SENDCMDOUTPARAMS* pOut = (SENDCMDOUTPARAMS*)(buffer + sizeof(SRB_IO_CONTROL));
        IDENTIFY_DATA*  pId = (IDENTIFY_DATA*)(pOut->bBuffer);
        if (pId->ModelNumber[0])
          print_ata_id(pId);
      }
    }
    ::CloseHandle(h);
  }
  else {
    res = ::GetLastError() == ERROR_ACCESS_DENIED;
    if (res)
      pc->PrintfFormattedString("%s, access denied\n", dev_file);
  }
  return res;
}

void main(){
  
  TConsoleIO console("NTIDEinfo v"VER_NUMBER_STR"."BUILD_NUMBER_STR" by Eugene Surovegin <ebs@innocent.com>");
  pc = &console;

  pc->WriteFormattedString("\n"LIGHTCYAN_B"NTIDEinfo v"VER_NUMBER_STR" build "BUILD_NUMBER_STR" ("__DATE__")\n");
  pc->WriteFormattedString(LIGHTCYAN_B"(C) 1999-2000 by Eugene Surovegin <ebs@innocent.com>\n");
  
  OSVERSIONINFO os = {0};
  os.dwOSVersionInfoSize = sizeof(os);
  ::GetVersionEx(&os);
  if ( os.dwPlatformId == VER_PLATFORM_WIN32_NT ){
    ::SetErrorMode(SEM_FAILCRITICALERRORS | SEM_NOGPFAULTERRORBOX | SEM_NOOPENFILEERRORBOX);
    int i = 0;
    while ( try_scsi_drive(i) ) ++i;
  }
  else
    pc->WriteFormattedString(LIGHTRED_B"Sorry, this utility for Windows NT ONLY\n");
}

#ifndef _DEBUG
extern "C" void mainCRTStartup(){
  main();
}
#endif